"""Utility functions.
"""
import functools
import operator
import itertools
import time
import math
import sys
from collections.abc import Iterable

import tqdm
import numpy as np


def isiterable(obj):
    return isinstance(obj, Iterable)


def prod(it):
    """Product of an iterable.
    """
    return functools.reduce(operator.mul, it)


def unzip(its, zip_level=1):
    """Split a nested iterable at a specified level, i.e. in numpy language
    transpose the specified 'axis' to be the first.

    Parameters
    ----------
    its: iterable (of iterables (of iterables ...))
        'n-dimensional' iterable to split
    zip_level: int
        level at which to split the iterable, default of 1 replicates
        ``zip(*its)`` behaviour.

    Example
    -------
    >>> x = [[(1, True), (2, False), (3, True)],
             [(7, True), (8, False), (9, True)]]
    >>> nums, bools = unzip(x, 2)
    >>> nums
    ((1, 2, 3), (7, 8, 9))
    >>> bools
    ((True, False, True), (True, False, True))

    """
    def _unzipper(its, zip_level):
        if zip_level > 1:
            return (zip(*_unzipper(it, zip_level - 1)) for it in its)
        else:
            return its

    return zip(*_unzipper(its, zip_level)) if zip_level else its


def flatten(its, n):
    """Take the n-dimensional nested iterable its and flatten it.

    Parameters
    ----------
        its : nested iterable
        n : number of dimensions

    Returns
    -------
        flattened iterable of all items
    """
    if n > 1:
        return itertools.chain(*(flatten(it, n - 1) for it in its))
    else:
        return its


def _get_fn_name(fn):
    """Try to inspect a function's name, taking into account several common
    non-standard types of function: dask, functools.partial ...
    """
    if hasattr(fn, "__name__"):
        return fn.__name__
    # try dask delayed function with key
    elif hasattr(fn, "key"):
        return fn.key.partition('-')[0]
    # try functools.partial function syntax
    elif hasattr(fn, "func"):
        return fn.func.__name__
    else:
        raise ValueError("Could not extract function name from {}".format(fn))


def progbar(it=None, nb=False, **kwargs):
    """Turn any iterable into a progress bar, with notebook option

    Parameters
    ----------
        it: iterable
            Iterable to wrap with progress bar
        nb: bool
            Whether  to display the notebook progress bar
        **kwargs: dict-like
            additional options to send to tqdm
    """
    defaults = {'ascii': True, 'smoothing': 0.0}
    # Overide defaults with custom kwargs
    settings = {**defaults, **kwargs}
    if nb:  # pragma: no cover
        return tqdm.tqdm_notebook(it, **settings)
    return tqdm.tqdm(it, **settings)


def getsizeof(obj):
    """Compute the real size of a python object. Taken from

    https://stackoverflow.com/a/30316760/5640201
    """
    import sys
    from types import ModuleType, FunctionType
    from gc import get_referents

    # Custom objects know their class.
    # Function objects seem to know way too much, including modules.
    # Exclude modules as well.
    excluded = type, ModuleType, FunctionType
    if isinstance(obj, excluded):
        raise TypeError(
            'getsize() does not take argument of type: {}'.format(type(obj)))

    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for obj in objects:
            if not isinstance(obj, excluded) and id(obj) not in seen_ids:
                seen_ids.add(id(obj))
                size += sys.getsizeof(obj)
                need_referents.append(obj)
        objects = get_referents(*need_referents)
    return size


class Timer:
    """A very simple context manager class for timing blocks.

    Examples
    --------

    >>> from xyzpy import Timer
    >>> with Timer() as timer:
    ...     print('Doing some work!')
    ...
    Doing some work!
    >>> timer.t
    0.00010752677917480469

    """

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.t = self.time = self.interval = self.end - self.start


def _auto_min_time(timer, min_t=0.2, repeats=5, get='min'):
    tot_t = 0
    number = 1

    while True:
        tot_t = timer.timeit(number)
        if tot_t > min_t:
            break
        number *= 2

    results = [tot_t] + timer.repeat(repeats - 1, number)

    if get == 'mean':
        return sum(results) / (number * len(results))

    return min(t / number for t in results)


def benchmark(fn, setup=None, n=None, min_t=0.1,
              repeats=3, get='min', starmap=False):
    """Benchmark the time it takes to run ``fn``.

    Parameters
    ----------
    fn : callable
        The function to time.
    setup : callable, optional
        If supplied the function that sets up the argument for ``fn``.
    n : int, optional
        If supplied, the integer to supply to ``setup`` of ``fn``.
    min_t : float, optional
        Aim to repeat function enough times to take up this many seconds.
    repeats : int, optional
        Repeat the whole procedure (with setup) this many times in order to
        take the minimum run time.
    get : {'min', 'mean'}, optional
        Return the minimum or mean time for each run.
    starmap : bool, optional
        Unpack the arguments from ``setup``, if given.

    Returns
    -------
    t : float
        The minimum, averaged, time to run ``fn`` in seconds.

    Examples
    --------

    Just a parameter-less function:

        >>> import xyzpy as xyz
        >>> import numpy as np
        >>> xyz.benchmark(lambda: np.linalg.eig(np.random.randn(100, 100)))
        0.004726233000837965

    The same but with a setup and size parameter ``n`` specified:

        >>> setup = lambda n: np.random.randn(n, n)
        >>> fn = lambda X: np.linalg.eig(X)
        >>> xyz.benchmark(fn, setup, 100)
        0.0042192734545096755
    """
    from timeit import Timer

    if n is None:
        n = ""

    if setup is None:
        setup_str = ""
        stmnt_str = "fn({})".format(n)
    else:
        setup_str = "X=setup({})".format(n)
        stmnt_str = "fn(*X)" if starmap else "fn(X)"

    timer = Timer(setup=setup_str, stmt=stmnt_str,
                  globals={'setup': setup, 'fn': fn})

    return _auto_min_time(timer, min_t=min_t, repeats=repeats, get=get)


class Benchmarker:
    """Compare the performance of various ``kernels``. Internally this makes
    use of :func:`~xyzpy.benchmark`, :func:`~xyzpy.Harvester` and xyzpys
    plotting functionality.

    Parameters
    ----------
    kernels : sequence of callable
        The functions to compare performance with.
    setup : callable, optional
        If given, setup each benchmark run by suppling the size argument ``n``
        to this function first, then feeding its output to each of the
        functions.
    names : sequence of str, optional
        Alternate names to give the function, else they will be inferred.
    benchmark_opts : dict, optional
        Supplied to :func:`~xyzpy.benchmark`.
    data_name : str, optional
        If given, the file name the internal harvester will use to store
        results persistently.

    Attributes
    ----------
    harvester : xyz.Harvester
        The harvester that runs and accumulates all the data.
    ds : xarray.Dataset
        Shortcut to the harvester's full dataset.
    """

    def __init__(self, kernels, setup=None, names=None,
                 benchmark_opts=None, data_name=None):
        import xyzpy as xyz

        self.kernels = kernels
        self.names = [f.__name__ for f in kernels] if names is None else names
        self.setup = setup
        self.benchmark_opts = {} if benchmark_opts is None else benchmark_opts

        def time(n, kernel):
            fn = self.kernels[self.names.index(kernel)]
            return xyz.benchmark(fn, self.setup, n, **self.benchmark_opts)

        self.runner = xyz.Runner(time, ['time'])
        self.harvester = xyz.Harvester(self.runner, data_name=data_name)

    def run(self, ns, kernels=None, **harvest_opts):
        """Run the benchmarks. Each run accumulates rather than overwriting the
        results.

        Parameters
        ----------
        ns : sequence of int or int
            The sizes to run the benchmarks with.
        kernels : sequence of str, optional
            If given, only run the kernels with these names.
        harvest_opts
            Supplied to :meth:`~xyzpy.Harvester.harvest_combos`.
        """
        if not isiterable(ns):
            ns = (ns,)

        if kernels is None:
            kernels = self.names

        combos = {'n': ns, 'kernel': kernels}
        self.harvester.harvest_combos(combos, **harvest_opts)

    @property
    def ds(self):
        return self.harvester.full_ds

    def lineplot(self, **plot_opts):
        """Plot the benchmarking results.
        """
        plot_opts.setdefault('xlog', True)
        plot_opts.setdefault('ylog', True)
        return self.ds.xyz.lineplot('n', 'time', 'kernel', **plot_opts)

    def ilineplot(self, **plot_opts):
        """Interactively plot the benchmarking results.
        """
        plot_opts.setdefault('xlog', True)
        plot_opts.setdefault('ylog', True)
        return self.ds.xyz.ilineplot('n', 'time', 'kernel', **plot_opts)


class RunningStatistics:  # pragma: no cover
    """Running mean & standard deviation using Welford's
    algorithm. This is a very efficient way of keeping track of the error on
    the mean for example.

    Attributes
    ----------
    mean : float
        Current mean.
    count : int
        Current count.
    std : float
        Current standard deviation.
    var : float
        Current variance.
    err : float
        Current error on the mean.
    rel_err: float
        The current relative error.

    Examples
    --------

        >>> rs = RunningStatistics()
        >>> rs.update(1.1)
        >>> rs.update(1.4)
        >>> rs.update(1.2)
        >>> rs.update_from_it([1.5, 1.3, 1.6])
        >>> rs.mean
        1.3499999046325684

        >>> rs.std  # standard deviation
        0.17078252585383266

        >>> rs.err  # error on the mean
        0.06972167422092768

    """

    def __init__(self):
        self.count = 0
        self.mean = 0.0
        self.M2 = 0.0

    def update(self, x):
        """Add a single value ``x`` to the statistics.
        """
        self.count += 1
        delta = x - self.mean
        self.mean += delta / self.count
        delta2 = x - self.mean
        self.M2 += delta * delta2

    def update_from_it(self, xs):
        """Add all values from iterable ``xs`` to the statistics.
        """
        for x in xs:
            self.update(x)

    def converged(self, rtol, atol):
        """Check if the stats have converged with respect to relative and
        absolute tolerance ``rtol`` and ``atol``.
        """
        return self.err < rtol * abs(self.mean) + atol

    @property
    def var(self):
        if self.count == 0:
            return np.inf
        return self.M2 / self.count

    @property
    def std(self):
        if self.count == 0:
            return np.inf
        return self.var**0.5

    @property
    def err(self):
        if self.count == 0:
            return np.inf
        return self.std / self.count**0.5

    @property
    def rel_err(self):
        if self.count == 0:
            return np.inf
        return self.err / abs(self.mean)


class RunningCovariance:  # pragma: no cover
    """Running covariance class.
    """

    def __init__(self):
        self.count = 0
        self.xmean = 0.0
        self.ymean = 0.0
        self.C = 0.0

    def update(self, x, y):
        self.count += 1
        dx = x - self.xmean
        dy = y - self.ymean
        self.xmean += dx / self.count
        self.ymean += dy / self.count
        self.C += dx * (y - self.ymean)

    def update_from_it(self, xs, ys):
        for x, y in zip(xs, ys):
            self.update(x, y)

    @property
    def covar(self):
        """The covariance.
        """
        return self.C / self.count

    @property
    def sample_covar(self):
        """The covariance with "Bessel's correction".
        """
        return self.C / (self.count - 1)


class RunningCovarianceMatrix:

    def __init__(self, n=2):
        self.n = n
        self.rcs = {}
        for i in range(self.n):
            for j in range(i, self.n):
                self.rcs[i, j] = RunningCovariance()

    def update(self, *x):
        for i in range(self.n):
            for j in range(i, self.n):
                self.rcs[i, j].update(x[i], x[j])

    def update_from_it(self, *xs):
        for i in range(self.n):
            for j in range(i, self.n):
                self.rcs[i, j].update_from_it(xs[i], xs[j])

    @property
    def count(self):
        return self.rcs[0, 0].count

    @property
    def covar_matrix(self):
        covar_matrix = np.empty((self.n, self.n))
        for i in range(self.n):
            for j in range(self.n):
                if j >= i:
                    covar_matrix[i, j] = self.rcs[i, j].covar
                else:
                    covar_matrix[i, j] = self.rcs[j, i].covar
        return covar_matrix

    @property
    def sample_covar_matrix(self):
        covar_matrix = np.empty((self.n, self.n))
        for i in range(self.n):
            for j in range(self.n):
                if j >= i:
                    covar_matrix[i, j] = self.rcs[i, j].sample_covar
                else:
                    covar_matrix[i, j] = self.rcs[j, i].sample_covar
        return covar_matrix

    def to_uncertainties(self, bias=True):
        """Convert the accumulated statistics to correlated uncertainties,
        from which new quantities can be calculated with error automatically
        propagated.

        Parameters
        ----------
        bias : bool, optional
            If False, use the sample covariance with "Bessel's correction".

        Return
        ------
        values : tuple of uncertainties.ufloat
            The sequence of correlated variables.

        Examples
        --------

        Estimate quantities of two perfectly correlated sequences.

            >>> rcm = xyz.RunningCovarianceMatrix()
            >>> rcm.update_from_it((1, 3, 2), (2, 6, 4))
            >>> x, y = rcm.to_uncertainties(rcm)

        Calculated quantities like sums have the error propagated:

            >>> x + y
            6.0+/-2.4494897427831783

        But the covariance is also taken into account, meaning the ratio here
        can be estimated with zero error:

            >>> x / y
            0.5+/-0

        """
        import uncertainties

        means = [self.rcs[i, i].xmean for i in range(self.n)]
        if bias:
            covar = self.covar_matrix
        else:
            covar = self.sample_covar_matrix

        return uncertainties.correlated_values(means, covar)


def estimate_from_repeats(fn, *fn_args, rtol=0.02, tol_scale=1.0, get='stats',
                          verbosity=0, min_samples=5, max_samples=1000000,
                          **fn_kwargs):
    """
    Parameters
    ----------
    fn : callable
        The function that estimates a single value.
    fn_args, optional
        Supplied to ``fn``.
    rtol : float, optional
        Relative tolerance for error on mean.
    tol_scale : float, optional
        The expected 'scale' of the estimate, this modifies the aboslute
        tolerance near zero to ``rtol * tol_scale``, default: 1.0.
    get : {'stats', 'samples', 'mean'}, optional
        Just get the ``RunningStatistics`` object, or the actual samples too,
        or just the actual mean estimate.
    verbosity : { 0, 1, 2}, optional
        How much information to show:

        - ``0``: nothing
        - ``1``: progress bar just with iteration rate
        - ``2``: progress bar with running stats displayed.

    min_samples : int, optional
        Take at least this many samples before checking for convergence.
    max_samples : int, optional
        Take at maximum this many samples.
    fn_kwargs, optional
        Supplied to ``fn``.

    Returns
    -------
    rs : RunningStatistics
        Statistics about the random estimation.
    samples : list[float]
        If ``get=='samples'``, the actual samples.


    Examples
    --------

    Estimate the sum of ``n`` random numbers:

        >>> import numpy as np
        >>> import xyzpy as xyz
        >>> def fn(n):
        ...     return np.random.rand(n).sum()
        ...
        >>> stats = xyz.estimate_from_repeats(fn, n=10, verbosity=3)
        51: mean: 5.002 err: 0.119: : 0it [00:00, ?it/s]
        <RunningStatistics(mean=5.002, err=0.119, count=51)>

    """

    rs = RunningStatistics()
    repeats = itertools.count()

    if verbosity >= 1:
        repeats = progbar(repeats)
        prec = abs(round(math.log10(tol_scale * rtol))) + 1

    if get == 'samples':
        xs = []

    try:
        for i in repeats:
            x = fn(*fn_args, **fn_kwargs)
            if get == 'samples':
                xs.append(x)
            rs.update(x)

            if verbosity >= 2:
                desc = '{}: mean: {:.{prec}f} err: {:.{prec}f}'
                repeats.set_description(
                    desc.format(rs.count, rs.mean, rs.err, prec=prec))

            # need at least min_samples to check convergence
            if (i > min_samples):
                if rs.converged(rtol, tol_scale * rtol):
                    break

            # reached the maximum number of samples to try
            if i >= max_samples - 1:
                break

    # allow user to cleanly interupt sampling with keyboard
    except KeyboardInterrupt:
        pass

    finally:
        if verbosity >= 1:
            repeats.close()

    if verbosity >= 1:
        sys.stderr.flush()
        print("<RunningStatistics(mean={:.{prec}f}, err={:.{prec}f}, "
              "count={})>".format(rs.mean, rs.err, rs.count, prec=prec))

    if get == 'samples':
        return rs, xs

    if get == 'mean':
        return rs.mean

    return rs



from collections import OrderedDict
from functools import partial
import pytest
import numpy as np
from numpy.testing import assert_allclose
import xarray as xr

from xyzpy.gen.combo_runner import (
    combo_runner,
    results_to_ds,
    combo_runner_to_ds,
)
from . import (
    foo3_scalar,
    foo3_float_bool,
    foo2_array,
    foo2_array_bool,
    foo2_array_array,
    foo2_zarray1_zarray2,
    foo_array_input,
)


# --------------------------------------------------------------------------- #
# COMBO_RUNNER tests                                                          #
# --------------------------------------------------------------------------- #

_test_combos1 = (('a', [1, 2]),
                 ('b', [10, 20, 30]),
                 ('c', [100, 200, 300, 400]))

_test_expect1 = (np.array([1, 2]).reshape((2, 1, 1)) +
                 np.array([10, 20, 30]).reshape((1, 3, 1)) +
                 np.array([100, 200, 300, 400]).reshape((1, 1, 4)))


class TestComboRunner:

    @pytest.mark.parametrize('shuffle', [False, True, 2])
    def test_simple(self, shuffle):
        x = combo_runner(foo3_scalar, _test_combos1, shuffle=shuffle)
        assert_allclose(x, _test_expect1)

    def test_progbars(self):
        combo_runner(foo3_scalar, _test_combos1, verbosity=2)

    def test_dict(self):
        combos = OrderedDict(_test_combos1)
        x = combo_runner(foo3_scalar, combos)
        assert_allclose(x, _test_expect1)

    def test_single_combo(self):
        combos = [('a', [1, 2])]
        x = combo_runner(partial(foo3_scalar, b=20, c=300), combos)
        assert_allclose(x, [321, 322])

    def test_single_combo_single_tuple(self):
        combos = ('a', [1, 2])
        constants = {'b': 20, 'c': 300}
        x = combo_runner(foo3_scalar, combos, constants=constants)
        assert_allclose(x, [321, 322])

    def test_multires(self):
        combos = _test_combos1
        x, y = combo_runner(foo3_float_bool, combos, split=True)
        xn = _test_expect1
        yn = (np.array([1, 2]).reshape((2, 1, 1)) %
              np.array([2] * 24).reshape((2, 3, 4))) == 0
        assert_allclose(x, xn)
        assert_allclose(y, yn)

    @pytest.mark.parametrize('shuffle', [False, True, 2])
    @pytest.mark.parametrize('parallel', [False, True])
    @pytest.mark.parametrize('fn', (foo3_scalar,))
    def test_parallel_basic(self, parallel, fn, shuffle):
        x = combo_runner(fn, _test_combos1, num_workers=2, parallel=parallel,
                         shuffle=shuffle)
        assert_allclose(x, _test_expect1)

    @pytest.mark.parametrize('executor', ['cf-process', 'cf-thread',
                                          'mp-process', 'mp-thread'])
    @pytest.mark.parametrize('fn', (foo3_scalar,))
    def test_executor_basic(self, executor, fn):
        import concurrent.futures as cf
        import multiprocessing as mp
        executor = {
            'cf-process': cf.ProcessPoolExecutor,
            'cf-thread': cf.ThreadPoolExecutor,
            'mp-process': mp.Pool,
            'mp-thread': mp.pool.ThreadPool,
        }[executor](2)
        x = combo_runner(fn, _test_combos1, executor=executor)
        assert_allclose(x, _test_expect1)

    @pytest.mark.parametrize('parallel', [False, True])
    def test_parallel_multires(self, parallel):
        x = combo_runner(foo3_float_bool, _test_combos1, num_workers=2,
                         split=True, parallel=parallel)
        assert_allclose(x[0], _test_expect1)
        assert np.all(np.asarray(x[1])[1, ...])

    @pytest.mark.parametrize('parallel', [False, True])
    def test_parallel_dict(self, parallel):
        combos = OrderedDict(_test_combos1)
        x = [*combo_runner(foo3_scalar, combos, num_workers=2,
                           parallel=parallel, verbosity=2)]
        assert_allclose(x, _test_expect1)


class TestCombosToDS:

    def test_simple(self):
        results = [1, 2, 3]
        combos = [('a', [1, 2, 3])]
        var_names = ['sum']
        ds = results_to_ds(results, combos, var_names,
                           var_dims={'sum': ()}, var_coords={})
        assert ds['sum'].data.dtype == int


class TestComboRunnerToDS:

    @pytest.mark.parametrize('shuffle', [False, True, 2])
    def test_basic(self, shuffle):
        combos = _test_combos1
        ds = combo_runner_to_ds(foo3_scalar, combos, var_names=['bananas'],
                                shuffle=shuffle)
        assert ds.sel(a=2, b=30, c=400)['bananas'].data == 432

    def test_multiresult(self):
        ds = combo_runner_to_ds(foo3_float_bool, _test_combos1,
                                var_names=['bananas', 'cakes'])
        assert ds.bananas.data.dtype == int
        assert ds.cakes.data.dtype == bool
        assert ds.sel(a=2, b=30, c=400)['bananas'].data == 432
        assert ds.sel(a=1, b=10, c=100)['bananas'].data == 111
        assert ds.sel(a=2, b=30, c=400)['cakes'].data
        assert not ds.sel(a=1, b=10, c=100)['cakes'].data

    def test_arrayresult(self):
        combos = (('a', [1, 2]),
                  ('b', [10, 20, 30]))
        ds = combo_runner_to_ds(foo2_array, combos,
                                var_names='bananas',
                                var_dims={'bananas': ['sugar']},
                                var_coords={'sugar': [*range(10)]})
        assert ds.bananas.data.dtype == float
        assert_allclose(ds.sel(a=2, b=30)['bananas'].data,
                        [32.0, 32.1, 32.2, 32.3, 32.4,
                         32.5, 32.6, 32.7, 32.8, 32.9])

    def test_array_and_single_result(self):
        combos = (('a', [1, 2]),
                  ('b', [10, 20, 30]))
        ds = combo_runner_to_ds(foo2_array_bool, combos,
                                var_names=['bananas', 'ripe'],
                                var_dims=(['sugar'], []),
                                var_coords={'sugar': [*range(10, 20)]})
        assert ds.ripe.data.dtype == bool
        assert ds.sel(a=2, b=30, sugar=14)['bananas'].data == 32.4
        with pytest.raises((KeyError, ValueError)):
            ds['ripe'].sel(sugar=12)

    def test_single_string_var_names_with_no_var_dims(self):
        combos = ('a', [1, 2, 3])
        ds = combo_runner_to_ds(foo3_scalar, combos,
                                constants={'b': 10, 'c': 100},
                                var_names='sum')
        assert_allclose(ds['sum'].data, np.array([111, 112, 113]))

    def test_double_array_return_with_same_dimensions(self):
        combos = (('a', [1, 2]),
                  ('b', [10, 20, 30]))
        ds = combo_runner_to_ds(foo2_array_array, combos,
                                var_names=['apples', 'oranges'],
                                var_dims={('apples', 'oranges'): ['seeds']},
                                var_coords={'seeds': [*range(5)]})
        assert ds.oranges.data.dtype == int
        assert_allclose(ds.sel(a=2, b=30).apples.data, [30, 32, 34, 36, 38])

        assert_allclose(ds.sel(a=2, b=30).oranges.data, [30, 28, 26, 24, 22])
        assert 'seeds' in ds.apples.coords
        assert 'seeds' in ds.oranges.coords

    def test_double_array_return_with_no_given_dimensions(self):
        ds = combo_runner_to_ds(foo2_array_array,
                                combos=[('a', [1, 2]),
                                        ('b', [30, 40])],
                                var_names=['array1', 'array2'],
                                var_dims=[['auto'], ['auto']])
        assert (ds['auto'].data.dtype == int or
                ds['auto'].data.dtype == np.int64)
        assert_allclose(ds['auto'].data, [0, 1, 2, 3, 4])

    def test_complex_output(self):
        ds = combo_runner_to_ds(foo2_zarray1_zarray2,
                                combos=[('a', [1, 2]),
                                        ('b', [30, 40])],
                                var_names=['array1', 'array2'],
                                var_dims=[['auto'], ['auto']])
        assert ds['array1'].data.size == 2 * 2 * 5
        assert ds['array2'].data.size == 2 * 2 * 5
        assert ds['array1'].data.dtype == complex
        assert ds['array2'].data.dtype == complex
        assert_allclose(ds['array1'].sel(a=2, b=30).data,
                        32 + np.arange(5) * 0.1j)
        assert_allclose(ds['array2'].sel(a=2, b=30).data,
                        32 - np.arange(5) * 0.1j)

    def test_constants_to_attrs(self):
        ds = combo_runner_to_ds(foo3_scalar,
                                combos=[('a', [1, 2, 3]),
                                        ('c', [100, 200, 300])],
                                constants={'b': 20},
                                var_names='x')
        assert ds.attrs['b'] == 20

    def test_const_array_to_coord(self):
        ds = combo_runner_to_ds(foo_array_input,
                                combos=[('a', [1, 2, 3])],
                                constants={'t': [10, 20, 30]},
                                var_names=['x'],
                                var_dims=[['t']])
        assert 't' in ds.dims
        assert 't' not in ds.attrs

    def test_when_results_are_xobjs(self):

        def fn_ds(a, b):
            ds = xr.Dataset(
                coords={
                    'x': [1, 2],
                    'y': ['foo', 'bar', 'baz'],
                },
                data_vars={
                    'apples': (['x', 'y'], np.tile(10 * a + b, (2, 3))),
                    'lemons': ('y', np.tile(10 * a + b, (3,))),
                }
            )
            return ds

        fds = combo_runner_to_ds(
            fn_ds,
            combos={
                'a': [2, 3, 4, 5],
                'b': [6, 7, 8, 9, 10],
            }, var_names=None)

        assert fds.sel(a=4, b=9, x=2, y='bar')['apples'].values == 49
        assert fds.sel(a=5, b=7, y='bar')['lemons'].values == 57
        assert 'output' not in fds



import xarray as xr
import numpy as np
from numpy.testing import assert_allclose

from xyzpy.gen.case_runner import (
    case_runner,
    case_runner_to_ds,
    case_runner_to_df,
    find_missing_cases,
)
from . import (
    foo3_scalar,
    foo3_float_bool,
    foo2_array,
    foo2_array_array,
    foo2_zarray1_zarray2,
)


# --------------------------------------------------------------------------- #
# CASE_RUNNER tests                                                           #
# --------------------------------------------------------------------------- #

class TestCaseRunner:
    def test_seq(self):
        cases = ((1, 10, 100),
                 (2, 20, 200),
                 (3, 30, 300))
        xs = case_runner(foo3_scalar, ('a', 'b', 'c'), cases, verbosity=0)
        assert xs == (111, 222, 333)

    def test_progbar(self):
        cases = ((1, 10, 100),
                 (2, 20, 200),
                 (3, 30, 300))
        xs = case_runner(foo3_scalar, ('a', 'b', 'c'), cases, verbosity=2)
        assert xs == (111, 222, 333)

    def test_constants(self):
        cases = ((1,),
                 (2,),
                 (3,))
        xs = case_runner(foo3_scalar, ('a', 'b', 'c'), cases,
                         constants={'b': 10, 'c': 100})
        assert xs == (111, 112, 113)

    def test_parallel(self):
        cases = ((1, 10, 100),
                 (2, 20, 200),
                 (3, 30, 300))
        xs = case_runner(foo3_scalar, ('a', 'b', 'c'), cases, num_workers=1)
        assert xs == (111, 222, 333)

    def test_split(self):
        cases = ((1, 10, 100),
                 (2, 20, 200),
                 (3, 30, 300))
        a, b = case_runner(foo3_float_bool, ('a', 'b', 'c'), cases, split=True)
        assert a == (111, 222, 333)
        assert b == (False, True, False)

    def test_single_args(self):
        cases = (1, 2, 3)
        xs = case_runner(foo3_scalar, 'a', cases,
                         constants={'b': 10, 'c': 100})
        assert xs == (111, 112, 113)


class TestCaseRunnerToDS:
    def test_single(self):
        cases = [(1, 20, 300),
                 (3, 20, 100)]
        ds = case_runner_to_ds(foo3_scalar, ['a', 'b', 'c'], cases=cases,
                               var_names='sum')
        assert_allclose(ds['a'].data, [1, 3])
        assert_allclose(ds['b'].data, [20])
        assert_allclose(ds['c'].data, [100, 300])
        assert ds['sum'].loc[{'a': 1, 'b': 20, 'c': 300}].data == 321
        assert ds['sum'].loc[{'a': 3, 'b': 20, 'c': 100}].data == 123
        assert ds['sum'].loc[{'a': 1, 'b': 20, 'c': 100}].isnull()
        assert ds['sum'].loc[{'a': 3, 'b': 20, 'c': 300}].isnull()

    def test_single_dict_cases(self):
        cases = [{'a': 1, 'b': 20, 'c': 300}, {'a': 3, 'b': 20, 'c': 100}]
        ds = case_runner_to_ds(foo3_scalar, None, cases=cases, var_names='sum')
        assert_allclose(ds['a'].data, [1, 3])
        assert_allclose(ds['b'].data, [20])
        assert_allclose(ds['c'].data, [100, 300])
        assert ds['sum'].loc[{'a': 1, 'b': 20, 'c': 300}].data == 321
        assert ds['sum'].loc[{'a': 3, 'b': 20, 'c': 100}].data == 123
        assert ds['sum'].loc[{'a': 1, 'b': 20, 'c': 100}].isnull()
        assert ds['sum'].loc[{'a': 3, 'b': 20, 'c': 300}].isnull()

    def test_multires(self):
        cases = [(1, 20, 300),
                 (3, 20, 100)]
        ds = case_runner_to_ds(foo3_float_bool,
                               fn_args=['a', 'b', 'c'],
                               cases=cases,
                               var_names=['sum', 'a_even'])
        assert_allclose(ds['a'].data, [1, 3])
        assert_allclose(ds['b'].data, [20])
        assert_allclose(ds['c'].data, [100, 300])
        assert ds['sum'].loc[{'a': 1, 'b': 20, 'c': 300}].data == 321
        assert ds['sum'].loc[{'a': 3, 'b': 20, 'c': 100}].data == 123
        assert ds['sum'].loc[{'a': 1, 'b': 20, 'c': 100}].isnull()
        assert ds['sum'].loc[{'a': 3, 'b': 20, 'c': 300}].isnull()
        assert ds['a_even'].data.dtype == object
        assert bool(ds['a_even'].sel(a=1, b=20, c=300).data) is False
        assert bool(ds['a_even'].sel(a=3, b=20, c=100).data) is False
        assert ds['a_even'].loc[{'a': 1, 'b': 20, 'c': 100}].isnull()
        assert ds['a_even'].loc[{'a': 3, 'b': 20, 'c': 300}].isnull()

    def test_array_return(self):
        ds = case_runner_to_ds(fn=foo2_array, fn_args=['a', 'b'],
                               cases=[(2, 30), (4, 50)],
                               var_names='x',
                               var_dims=['time'],
                               var_coords={'time': np.arange(10) / 10})
        assert ds.x.data.dtype == float
        assert ds.x.sel(a=2, b=50, time=0.7).isnull()
        assert ds.x.sel(a=4, b=50, time=0.3).data == 54.3

    def test_multi_array_return(self):
        ds = case_runner_to_ds(fn=foo2_array_array, fn_args=['a', 'b'],
                               cases=[(2, 30), (4, 50)],
                               var_names=['x', 'y'],
                               var_dims={('x', 'y'): 'time'},
                               var_coords={'time': ['a', 'b', 'c', 'd', 'e']})
        assert ds['time'].data.dtype != object
        assert_allclose(ds['x'].sel(a=4, b=50).data,
                        [50, 54, 58, 62, 66])
        assert_allclose(ds['y'].sel(a=4, b=50).data,
                        [50, 46, 42, 38, 34])

    def test_align_and_fillna_int(self):
        ds1 = case_runner_to_ds(foo2_array_array, fn_args=['a', 'b'],
                                cases=[(1, 10), (2, 20)],
                                var_names=['x', 'y'],
                                var_dims={('x', 'y'): 'time'},
                                var_coords={'time':
                                            ['a', 'b', 'c', 'd', 'e']})
        ds2 = case_runner_to_ds(foo2_array_array, fn_args=['a', 'b'],
                                cases=[(2, 10), (1, 20)],
                                var_names=['x', 'y'],
                                var_dims={('x', 'y'): 'time'},
                                var_coords={'time':
                                            ['a', 'b', 'c', 'd', 'e']})
        assert not np.logical_not(ds1['x'].isnull()).all()
        assert not np.logical_not(ds1['y'].isnull()).all()
        assert not np.logical_not(ds2['x'].isnull()).all()
        assert not np.logical_not(ds2['y'].isnull()).all()
        ds1, ds2 = xr.align(ds1, ds2, join='outer')
        fds = ds1.fillna(ds2)
        assert np.logical_not(fds['x'].isnull()).all()
        assert np.logical_not(fds['y'].isnull()).all()

    def test_align_and_fillna_complex(self):
        ds1 = case_runner_to_ds(foo2_zarray1_zarray2, fn_args=['a', 'b'],
                                cases=[(1j, 10), (2j, 20)],
                                var_names=['x', 'y'],
                                var_dims={('x', 'y'): 'time'},
                                var_coords={'time':
                                            ['a', 'b', 'c', 'd', 'e']})
        ds2 = case_runner_to_ds(foo2_zarray1_zarray2, fn_args=['a', 'b'],
                                cases=[(2j, 10), (1j, 20)],
                                var_names=['x', 'y'],
                                var_dims={('x', 'y'): 'time'},
                                var_coords={'time':
                                            ['a', 'b', 'c', 'd', 'e']})
        assert not np.logical_not(np.isnan(ds1['x'].data)).all()
        assert not np.logical_not(np.isnan(ds1['y'].data)).all()
        assert not np.logical_not(np.isnan(ds2['x'].data)).all()
        assert not np.logical_not(np.isnan(ds2['y'].data)).all()
        assert all(t == complex for t in (ds1.x.dtype, ds2.x.dtype,
                                          ds1.y.dtype, ds2.y.dtype))
        assert ds1.y.dtype == complex
        assert ds2.y.dtype == complex
        ds1, ds2 = xr.align(ds1, ds2, join='outer')
        fds = ds1.fillna(ds2)
        assert np.logical_not(np.isnan(fds['x'].data)).all()
        assert np.logical_not(np.isnan(fds['y'].data)).all()


class TestCaseRunnerToDF:

    def test_single_arg(self):

        df = case_runner_to_df(
            fn=foo3_scalar,
            fn_args=['a', 'b', 'c'],
            cases=[(1, 10, 100), (2, 20, 200)],
            var_names='x',
        )
        assert len(df) == 2
        assert df.columns.tolist() == ['a', 'b', 'c', 'x']
        assert 'int' in df.x.dtype.name


# --------------------------------------------------------------------------- #
# Finding and filling missing data                                            #
# --------------------------------------------------------------------------- #

class TestFindMissingCases:
    def test_simple(self):
        ds = xr.Dataset(coords={'a': [1, 2, 3], 'b': [40, 50]})
        ds['x'] = (('a', 'b'), np.array([[0.1, np.nan],
                                         [np.nan, 0.2],
                                         [np.nan, np.nan]]))
        # Target cases and settings
        t_cases = ((1, 50), (2, 40), (3, 40), (3, 50))
        t_configs = tuple(dict(zip(['a', 'b'], t_case)) for t_case in t_cases)
        # Missing cases and settings
        m_args, m_cases = find_missing_cases(ds)
        m_configs = tuple(dict(zip(m_args, m_case)) for m_case in m_cases)
        # Assert same set of coordinates
        assert all(t_config in m_configs for t_config in t_configs)
        assert all(m_config in t_configs for m_config in m_configs)

    def test_multires(self):
        ds = xr.Dataset(coords={'a': [1, 2, 3], 'b': [40, 50]})
        ds['x'] = (('a', 'b'), np.array([[0.1, np.nan],
                                         [np.nan, 0.2],
                                         [np.nan, np.nan]]))
        ds['y'] = (('a', 'b'), np.array([['a', None],
                                         [None, 'b'],
                                         [None, None]]))
        # Target cases and settings
        t_cases = ((1, 50), (2, 40), (3, 40), (3, 50))
        t_configs = tuple(dict(zip(['a', 'b'], t_case)) for t_case in t_cases)
        # Missing cases and settings
        m_args, m_cases = find_missing_cases(ds)
        m_configs = tuple(dict(zip(m_args, m_case)) for m_case in m_cases)
        # Assert same set of coordinates
        assert set(m_args) == {'a', 'b'}
        assert all(t_config in m_configs for t_config in t_configs)
        assert all(m_config in t_configs for m_config in m_configs)

    def test_var_dims_leave(self):
        ds = xr.Dataset(coords={'a': [1, 2, 3], 'b': [40, 50],
                                't': [0.1, 0.2, 0.3]})
        ds['x'] = (('a', 'b'), np.array([[0.1, np.nan],
                                         [np.nan, 0.2],
                                         [np.nan, np.nan]]))
        ds['y'] = (('a', 'b', 't'), np.array([[[0.2] * 3, [np.nan] * 3],
                                              [[np.nan] * 3, [0.4] * 3],
                                              [[np.nan] * 3, [np.nan] * 3]]))
        # Target cases and settings
        t_cases = ((1, 50), (2, 40), (3, 40), (3, 50))
        t_configs = tuple(dict(zip(['a', 'b'], t_case)) for t_case in t_cases)
        # Missing cases and settings
        m_args, m_cases = find_missing_cases(ds, ignore_dims='t')
        m_configs = tuple(dict(zip(m_args, m_case)) for m_case in m_cases)
        # Assert same set of coordinates
        assert set(m_args) == {'a', 'b'}
        assert all(t_config in m_configs for t_config in t_configs)
        assert all(m_config in t_configs for m_config in m_configs)



import os
from tempfile import TemporaryDirectory

import pytest
import numpy as np
import xarray as xr
from numpy.testing import assert_allclose

from xyzpy import combo_runner, combo_runner_to_ds, Runner, Harvester, label
from xyzpy.gen.cropping import (
    XYZError,
    Crop,
    parse_crop_details,
    grow,
    load_crops,
)

from . import (
    foo3_scalar,
    foo2_scalar,
    foo2_array,
    foo2_array_bool,
    foo2_dataset,
)


def foo_add(a, b, c):
    return a + b


class TestSowerReaper:
    @pytest.mark.parametrize(
        "fn, crop_name, crop_loc, expected",
        [
            (foo_add, None, None, '.xyz-foo_add'),
            (None, 'custom', None, '.xyz-custom'),
            (foo_add, 'custom', None, '.xyz-custom'),
            (foo_add, None, 'custom_dir', os.path.join('custom_dir',
                                                       '.xyz-foo_add')),
            (None, 'custom', 'custom_dir', os.path.join('custom_dir',
                                                        '.xyz-custom')),
            (foo_add, 'custom', 'custom_dir', os.path.join('custom_dir',
                                                           '.xyz-custom')),
            (None, None, None, 'raises'),
            (None, None, 'custom_dir', 'raises'),

        ])
    def test_parse_crop_details(self, fn, crop_name, crop_loc, expected):

        if expected == 'raises':
            with pytest.raises(ValueError):
                parse_crop_details(fn, crop_name, crop_loc)
        else:
            crop_loc = parse_crop_details(fn, crop_name, crop_loc)[0]
            assert crop_loc[-len(expected):] == expected

    def test_checks(self):
        with pytest.raises(ValueError):
            Crop(name='custom', save_fn=True)

        with pytest.raises(TypeError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=0.5)
            c.choose_batch_settings(combos=[('a', [1, 2])])

        with pytest.raises(ValueError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=-1)
            c.choose_batch_settings(combos=[('a', [1, 2])])

        with pytest.raises(ValueError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=1, num_batches=2)
            c.choose_batch_settings(combos=[('a', [1, 2, 3])])

        with pytest.raises(ValueError):
            c = Crop(fn=foo_add, save_fn=False, batchsize=2, num_batches=3)
            c.choose_batch_settings(combos=[('a', [1, 2, 3])])

        c = Crop(fn=foo_add, save_fn=False, batchsize=1, num_batches=3)
        c.choose_batch_settings(combos=[('a', [1, 2, 3])])

        c = Crop(fn=foo_add, save_fn=False, batchsize=2, num_batches=2)
        c.choose_batch_settings(combos=[('a', [1, 2, 3])])

        c = Crop(fn=foo_add, save_fn=False, batchsize=3, num_batches=1)
        c.choose_batch_settings(combos=[('a', [1, 2, 3])])

        with pytest.raises(XYZError):
            grow(1)

        print(c)
        repr(c)

    @pytest.mark.parametrize('shuffle', [False, True, 2])
    def test_batch(self, shuffle):

        combos = [
            ('a', [10, 20, 30]),
            ('b', [4, 5, 6, 7]),
        ]
        expected = combo_runner(foo_add, combos, constants={'c': True})

        with TemporaryDirectory() as tdir:

            # sow seeds
            crop = Crop(fn=foo_add, parent_dir=tdir, batchsize=5)

            assert not crop.is_prepared()
            assert crop.num_sown_batches == crop.num_results == -1

            crop.sow_combos(combos, constants={'c': True}, shuffle=shuffle)

            assert crop.is_prepared()
            assert crop.num_sown_batches == 3
            assert crop.num_results == 0

            # grow seeds
            for i in range(1, 4):
                grow(i, Crop(parent_dir=tdir, name='foo_add'))

                if i == 1:
                    assert crop.missing_results() == (2, 3,)

                    with pytest.raises(XYZError):
                        crop.reap()

            assert crop.is_ready_to_reap()
            assert not crop.check_bad()
            # reap results
            results = crop.reap()

        assert results == expected

    def test_field_name_and_overlapping(self):
        combos1 = [('a', [10, 20, 30]),
                   ('b', [4, 5, 6, 7])]
        expected1 = combo_runner(foo_add, combos1, constants={'c': True})

        combos2 = [('a', [40, 50, 60]),
                   ('b', [4, 5, 6, 7])]
        expected2 = combo_runner(foo_add, combos2, constants={'c': True})

        with TemporaryDirectory() as tdir:
            # sow seeds
            c1 = Crop(name='run1', fn=foo_add, parent_dir=tdir, batchsize=5)
            c1.sow_combos(combos1, constants={'c': True})
            c2 = Crop(name='run2', fn=foo_add, parent_dir=tdir, batchsize=5)
            c2.sow_combos(combos2, constants={'c': True})

            # grow seeds
            for i in range(1, 4):
                grow(i, Crop(parent_dir=tdir, name='run1'))
                grow(i, Crop(parent_dir=tdir, name='run2'))

            # reap results
            assert not c1.check_bad()
            assert not c2.check_bad()
            results1 = c1.reap()
            results2 = c2.reap()

        assert results1 == expected1
        assert results2 == expected2

    @pytest.mark.parametrize("num_workers", [None, 2])
    @pytest.mark.parametrize('shuffle', [False, True, 2])
    def test_crop_grow_missing(self, num_workers, shuffle):
        combos1 = [('a', [10, 20, 30]),
                   ('b', [4, 5, 6, 7])]
        expected1 = combo_runner(foo_add, combos1, constants={'c': True})
        with TemporaryDirectory() as tdir:
            c1 = Crop(name='run1', fn=foo_add, parent_dir=tdir, batchsize=5)
            c1.sow_combos(combos1, constants={'c': True}, shuffle=shuffle)
            c1.grow_missing(num_workers=num_workers)
            results1 = c1.reap()
        assert results1 == expected1

    def test_combo_reaper_to_ds(self):
        combos = (('a', [1, 2]),
                  ('b', [10, 20, 30]),
                  ('c', [100, 200, 300, 400]))

        with TemporaryDirectory() as tdir:

            # sow seeds
            crop = Crop(fn=foo3_scalar, parent_dir=tdir, batchsize=5)
            crop.sow_combos(combos)

            # check on disk repr works
            print(crop)
            repr(crop)

            # grow seeds
            for i in range(1, 6):
                crop.grow(i)

                if i == 3:
                    with pytest.raises(XYZError):
                        crop.reap_combos_to_ds(var_names=['bananas'])

            ds = crop.reap_combos_to_ds(var_names=['bananas'])

        assert ds.sel(a=2, b=30, c=400)['bananas'].data == 432

    @pytest.mark.parametrize("num_batches", [67, 98])
    def test_num_batches_doesnt_divide(self, num_batches):
        combos = (('a', [1, 2, 3]),
                  ('b', [10, 20, 30]),
                  ('c', range(100, 1101, 100)))

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir, num_batches=num_batches)
            crop.sow_combos(combos)
            assert crop.num_batches == num_batches
            crop.grow_missing()
            ds = crop.reap_combos_to_ds(var_names=['sum'])

        assert ds['sum'].sel(a=3, b=30, c=1100).data == 33

    @pytest.mark.parametrize('fn', [
        foo2_scalar,
        foo2_array,
        foo2_array_bool,
        foo2_dataset,
    ])
    def test_all_nan_result(self, fn):

        combos = (('a', [1, 2, 3]),
                  ('b', [10, 20, 30]))

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=fn, parent_dir=tdir)
            crop.sow_combos(combos)

            with pytest.raises(XYZError):
                crop.all_nan_result

            crop.grow(1)
            nres = crop.all_nan_result

            if fn is foo2_array_bool:
                assert len(nres) == 2
                assert_allclose(nres[0], np.broadcast_to(np.nan, [10]))
                assert_allclose(nres[1], np.broadcast_to(np.nan, []))

            if fn is foo2_dataset:
                ds_exp = xr.Dataset({'x': (['t1', 't2'],
                                           np.tile(np.nan, (2, 3)))})
                assert nres.identical(ds_exp)

    @pytest.mark.parametrize('shuffle', [False, True, 2])
    def test_reap_allow_incomplete(self, shuffle):
        combos = (('a', [1, 2, 3]),
                  ('b', [10, 20, 30]),
                  ('c', range(100, 1101, 100)))

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=foo_add, parent_dir=tdir)
            crop.sow_combos(combos, shuffle=shuffle)
            with pytest.raises(XYZError):
                crop.reap(allow_incomplete=True)
            for i in range(1, 40):
                crop.grow(i)
            res = np.array(crop.reap(allow_incomplete=True))
            assert np.isnan(res).sum() == 60

    @pytest.mark.parametrize(
        "fn,var_names,var_dims", [
            (foo2_scalar, ['x'], None),
            (foo2_array, ['x'], {'x': 't'}),
            (foo2_array_bool, ['x', 'y'], {'x': 't'}),
            (foo2_dataset, None, None),
        ]
    )
    def test_reap_to_ds_allow_incomplete(self, fn, var_names, var_dims):
        combos = (('a', [1, 2, 3]),
                  ('b', [10, 20, 30]))

        ds_exp = combo_runner_to_ds(fn, combos, var_names, var_dims=var_dims)

        with TemporaryDirectory() as tdir:
            crop = Crop(fn=fn, parent_dir=tdir)
            crop.sow_combos(combos)
            for i in range(1, 10, 2):
                crop.grow(i)

            ds = crop.reap_combos_to_ds(var_names=var_names, var_dims=var_dims,
                                        allow_incomplete=True)

            num_finished = int(ds['x'].size * (5 / 9))
            assert (ds['x'] == ds_exp['x']).sum() == num_finished

            crop.grow_missing()
            ds = crop.reap_combos_to_ds(var_names=var_names, var_dims=var_dims,
                                        allow_incomplete=True)
            assert ds.identical(ds_exp)

    def test_new_ds_crop_loads_info_incomplete(self):
        def fn(a, b):
            return xr.Dataset({'sum': a + b, 'diff': a - b})

        with TemporaryDirectory() as tdir:
            disk_ds = os.path.join(tdir, 'test.h5')

            combos = dict(a=[1], b=[1, 2, 3])
            runner = Runner(fn, var_names=None)
            harvester = Harvester(runner, disk_ds)
            crop = harvester.Crop(name='fn', batchsize=1, parent_dir=tdir)
            crop.sow_combos(combos)
            for i in range(1, 3):
                crop.grow(i)

            # try creating crop from fresh
            c = Crop(name='fn', parent_dir=tdir)
            # crop's harvester should be loaded from disk
            assert c.farmer is not None
            assert c.farmer is not harvester
            ds = c.reap(allow_incomplete=True)
            assert isinstance(ds, xr.Dataset)
            assert ds['diff'].isnull().sum() == 1
            assert harvester.full_ds['diff'].isnull().sum() == 1

            # try creating crop from harvester
            c = harvester.Crop('fn', parent_dir=tdir)
            # crop's harvester should still be harvester
            assert c.farmer is not None
            assert c.farmer is harvester
            ds = c.reap(allow_incomplete=True)
            assert isinstance(ds, xr.Dataset)
            assert ds['diff'].isnull().sum() == 1

    def test_load_crops(self):

        combos = (('a', [1, 2, 3]),
                  ('b', [10, 20, 30]),
                  ('c', range(100, 1101, 100)))

        with TemporaryDirectory() as tdir:

            c1 = Crop(name='Alice', fn=foo_add, parent_dir=tdir)
            c2 = Crop(name='Bob', fn=foo_add, parent_dir=tdir)

            c1.sow_combos(combos)
            c2.sow_combos(combos)

            crops = load_crops(tdir)
            assert 'Alice' in crops
            assert 'Bob' in crops
            assert len(crops) == 2

            c1.grow_missing()
            c2.grow_missing()

            assert (c1.reap_combos() ==
                    c2.reap_combos() ==
                    combo_runner(foo_add, combos))

    @pytest.mark.parametrize("scheduler", ['sge', 'pbs', 'slurm'])
    def test_gen_cluster_script(self, scheduler):
        combos = [
            ('a', [10, 20, 30]),
            ('b', [4, 5, 6, 7]),
        ]

        with TemporaryDirectory() as tdir:

            # sow seeds
            crop = Crop(fn=foo_add, parent_dir=tdir)
            crop.sow_combos(combos, constants={'c': True})

            # test script to grow all
            s1 = crop.gen_cluster_script(scheduler=scheduler, minutes=20)
            print(s1)

            # test script to grow specified
            s2 = crop.gen_cluster_script(batch_ids=[0, 1], scheduler=scheduler)
            print(s2)

            # test script to grow missing
            crop.grow((1, 3, 5))
            s3 = crop.gen_cluster_script(scheduler=scheduler)
            print(s3)

            assert s1 != s2
            assert s1 != s3
            assert s2 != s3

    def test_sow_reap_cases(self):

        def dummy_function(a, b):
            return a + b

        cases1 = [(1, 2), (3, 4)]
        cases2 = [(5, 6), (7, 8)]
        fn_args = ('a', 'b')

        with TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.dmp')
            runner = Runner(dummy_function, var_names=['Dummy_Value'])
            harvester = Harvester(runner, engine='joblib', data_name=fl_pth)

            crop1 = harvester.Crop(parent_dir=tmpdir)
            crop1.sow_cases(fn_args=fn_args, cases=cases1)
            crop1.grow_missing()
            crop1.reap()

            crop2 = harvester.Crop(parent_dir=tmpdir)
            crop2.sow_cases(fn_args=fn_args, cases=cases2)
            crop2.grow_missing()
            crop2.reap()

            ds = harvester.full_ds
            assert isinstance(ds, xr.Dataset)
            assert ds['Dummy_Value'].size == 16
            assert ds['Dummy_Value'].notnull().sum() == 4

    def test_sow_reap_mixed_combos_cases(self):

        @label('out')
        def fn(a, b, c, d, e):
            return f"{a}{b}{c}{d}{e}"

        with TemporaryDirectory() as tmpdir:
            crop = fn.Crop('test', parent_dir=tmpdir)
            cases = [
                {'a': 1, 'c': 3},
                {'a': 2, 'c': 4},
            ]
            combos = {
                'b': [5, 6],
                'd': [7, 8],
            }
            constants = {'e': 9}
            crop.sow_combos(combos=combos, cases=cases, constants=constants)
            crop.grow_missing()
            ds = crop.reap()

        assert ds['out'].ndim == 4
        assert ds['out'].notnull().sum().item() == 8



import os
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor

import pytest
import numpy as np
import xarray as xr

from xyzpy.manage import load_ds, load_df
from xyzpy.gen.farming import Runner, Harvester, Sampler, label
from xyzpy.gen.cropping import grow


# -------------------------------- Fixtures --------------------------------- #

def fn1_x(x):
    return x ** 2


def fn3_fba(a, b, c):
    sm = a + b + c
    ev = ((a + b + c) % 2 == 0)
    ts = a * (b * np.linspace(0, 1.0, 3) + c)
    return sm, int(ev), ts


@pytest.fixture
def fn3_fba_runner():
    r = Runner(fn3_fba, fn_args=('a', 'b'),
               var_names=['sum', 'even', 'array'],
               var_dims={'array': ['time']},
               var_coords={'time': np.linspace(0, 1.0, 3)},
               constants={'c': 100},
               attrs={'fruit': 'apples'})
    return r


@pytest.fixture
def fn3_fba_ds():
    return xr.Dataset(
        coords={'a': [1, 2], 'b': [3, 4], 'time': np.linspace(0, 1.0, 3)},
        data_vars={'sum': (('a', 'b'), [[104, 105],
                                        [105, 106]]),
                   'even': (('a', 'b'), [[True, False],
                                         [False, True]]),
                   'array': (('a', 'b', 'time'),
                             [[[100, 101.5, 103], [100, 102, 104]],
                              [[200, 203, 206], [200, 204, 208]]])},
        attrs={'c': 100, 'fruit': 'apples'})


# ------------------------------ Runner Tests ------------------------------- #

class TestRunner:

    def test_basic(self):
        r = Runner(fn1_x, "x^2")
        print(str(r))
        print(repr(r))
        assert r(2) == 4

    @pytest.mark.parametrize("prop, vals", [('var_names', ['S', 'E', 'A']),
                                            ('fn_args', ['A', 'B', 'C']),
                                            ('var_dims', {'array': ['t']}),
                                            ('var_coords', {'time': [1, 2]}),
                                            ('constants', [('c', 200)]),
                                            ('resources', [('c', 300)])])
    def test_properties(self, prop, vals):
        r = Runner(fn3_fba, fn_args=('a', 'b', 'c'),
                   var_names=['sum', 'even', 'array'],
                   var_dims={'array': ['time']},
                   var_coords={'time': np.linspace(0, 1.0, 3)},
                   constants={'c': 100},
                   attrs={'fruit': 'apples'})
        getattr(r, prop)
        setattr(r, prop, vals)
        delattr(r, prop)
        assert getattr(r, prop) is None

    @pytest.mark.parametrize("var_names", ['sq', ('sq',), ['sq']])
    @pytest.mark.parametrize("var_dims", [[], (), {}, {'sq': ()}])
    @pytest.mark.parametrize("combos", [('x', (1, 2, 3)),
                                        (('x', (1, 2, 3)),),
                                        ('x', [1, 2, 3]),
                                        [('x', [1, 2, 3])],
                                        {'x': (1, 2, 3)},
                                        {'x': [1, 2, 3]}])
    def test_runner_combos_parse(self, var_names, var_dims, combos):
        r = Runner(fn1_x, var_names=var_names, var_dims=var_dims)
        r.run_combos(combos)
        expected_ds = xr.Dataset(coords={'x': [1, 2, 3]},
                                 data_vars={'sq': ('x', [1, 4, 9])})
        assert r.last_ds.identical(expected_ds)

    def test_runner_combos(self, fn3_fba_runner, fn3_fba_ds):
        r = fn3_fba_runner
        r.run_combos((('a', (1, 2)), ('b', (3, 4))))
        assert r.last_ds.identical(fn3_fba_ds)

    def test_sow_reap_seperate(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            r = fn3_fba_runner
            crop = r.Crop(parent_dir=tmpdir, num_batches=2)
            crop.sow_combos((('a', (1, 2)),
                             ('b', (3, 4))))
            for i in [1, 2]:
                grow(i, crop)
            crop.reap()
            assert r.last_ds.identical(fn3_fba_ds)

    def test_sow_and_reap(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            r = fn3_fba_runner
            crop = r.Crop(parent_dir=tmpdir, num_batches=2)

            def concurrent_grow():
                # wait for cases to be sown
                time.sleep(0.5)
                for i in [1, 2]:
                    grow(i, crop)

            with ThreadPoolExecutor(1) as pool:
                pool.submit(concurrent_grow)

                crop.sow_combos((('a', (1, 2)),
                                 ('b', (3, 4))))
                crop.reap(wait=True)

            assert r.last_ds.identical(fn3_fba_ds)

    @pytest.mark.parametrize("cases", [(1, 2, 3),
                                       ([1], [2], [3])])
    def test_runner_cases_parse(self, cases):
        r = Runner(fn1_x, var_names='sq', fn_args='x')
        r.run_cases(cases)
        expected_ds = xr.Dataset(coords={'x': [1, 2, 3]},
                                 data_vars={'sq': ('x', [1, 4, 9])})
        assert r.last_ds.identical(expected_ds)

    @pytest.mark.parametrize("dict_cases", [False, True])
    def test_runner_cases(self, fn3_fba_runner, dict_cases):

        if dict_cases:
            cases = [{'a': 2, 'b': 3}, {'a': 1, 'b': 4}]
        else:
            cases = [(2, 3), (1, 4)]

        fn3_fba_runner.run_cases(cases)
        expected_ds = xr.Dataset(
            coords={'a': [1, 2], 'b': [3, 4], 'time': np.linspace(0, 1.0, 3)},
            data_vars={'sum': (('a', 'b'), [[np.nan, 105],
                                            [105, np.nan]]),
                       'even': (('a', 'b'), [[None, False],
                                             [False, None]]),
                       'array': (('a', 'b', 'time'),
                                 [[[np.nan, np.nan, np.nan],
                                   [100, 102, 104]],
                                  [[200, 203, 206],
                                   [np.nan, np.nan, np.nan]]])},
            attrs={'c': 100, 'fruit': 'apples'})
        assert fn3_fba_runner.last_ds.identical(expected_ds)


# -------------------------- Harvester Tests -------------------------------- #

class TestHarvester:

    def test_save_and_load_ds(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth, full_ds=fn3_fba_ds)
            h.save_full_ds()
            h.load_full_ds()
            assert h.full_ds.equals(fn3_fba_ds)

    def test_harvest_combos_new(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_combos((('a', (1, 2)), ('b', (3, 4))))
            hds = load_ds(fl_pth)
        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_label_as_harvester(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = label(
                harvester=fl_pth,
                fn_args=('a', 'b'),
                var_names=['sum', 'even', 'array'],
                var_dims={'array': ['time']},
                var_coords={'time': np.linspace(0, 1.0, 3)},
                constants={'c': 100},
                attrs={'fruit': 'apples'},
            )(fn3_fba)
            h.harvest_combos((('a', (1, 2)), ('b', (3, 4))))
            hds = load_ds(fl_pth)
        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_new_sow_reap_separate(self, fn3_fba_runner,
                                                  fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            crop = h.Crop(parent_dir=tmpdir, num_batches=2)

            crop.sow_combos((('a', (1, 2)), ('b', (3, 4))))

            for i in [1, 2]:
                grow(i, crop)
            crop.reap()

            hds = load_ds(fl_pth)

            assert h.last_ds.identical(fn3_fba_ds)
            assert h.full_ds.identical(fn3_fba_ds)
            assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_sow_and_reap_thread_wait(self, fn3_fba_runner,
                                                     fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            crop = h.Crop(parent_dir=tmpdir, num_batches=2)

            def concurrent_grow():
                # wait for cases to be sown
                time.sleep(0.5)
                for i in [1, 2]:
                    grow(i, crop)

            with ThreadPoolExecutor(1) as pool:
                pool.submit(concurrent_grow)
                crop.sow_combos((('a', (1, 2)), ('b', (3, 4))))
                crop.reap(wait=True)

            hds = load_ds(fl_pth)

        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_new_sow_reap_incomplete(self, fn3_fba_runner,
                                                    fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            crop = h.Crop(parent_dir=tmpdir, num_batches=3)
            crop.sow_combos((('a', (1, 2)), ('b', (3, 4))))

            grow(1, crop)
            crop.reap(allow_incomplete=True)
            assert not h.full_ds.identical(fn3_fba_ds)
            hds = load_ds(fl_pth)
            assert not hds.identical(fn3_fba_ds)

            crop.grow_missing()
            crop.reap(allow_incomplete=True)
            h.load_full_ds()
            assert h.full_ds.equals(fn3_fba_ds)
            hds = load_ds(fl_pth)
            assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_merge(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_combos((('a', (1,)), ('b', (3, 4))))
            h.harvest_combos((('a', (2,)), ('b', (3, 4))))
            hds = load_ds(fl_pth)
        assert not h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_combos_overwrite(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            mod_ds = fn3_fba_ds.copy(deep=True)
            mod_ds['array'].loc[{'a': 1, 'b': 3}] = 999
            h = Harvester(fn3_fba_runner, fl_pth, full_ds=mod_ds)
            h.save_full_ds()
            assert not h.full_ds.equals(fn3_fba_ds)
            h.harvest_combos((('a', (1,)), ('b', (3,))), overwrite=True)
            assert h.full_ds.equals(fn3_fba_ds)

    @pytest.mark.parametrize('dict_cases', [False, True])
    def test_harvest_cases_new(self, fn3_fba_runner, fn3_fba_ds, dict_cases):

        if dict_cases:
            cases = [{'a': 1, 'b': 3}, {'a': 1, 'b': 4},
                     {'a': 2, 'b': 3}, {'a': 2, 'b': 4}]
        else:
            cases = [(1, 3), (1, 4), (2, 3), (2, 4)]

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_cases(cases)
            hds = load_ds(fl_pth)
        assert h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_cases_merge(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth)
            h.harvest_cases([(1, 3), (2, 4)])
            h.harvest_cases([(1, 4), (2, 3)])
            hds = load_ds(fl_pth)
        assert not h.last_ds.identical(fn3_fba_ds)
        assert h.full_ds.identical(fn3_fba_ds)
        assert hds.identical(fn3_fba_ds)

    def test_harvest_cases_overwrite(self, fn3_fba_runner, fn3_fba_ds):
        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            mod_ds = fn3_fba_ds.copy(deep=True)
            mod_ds['array'].loc[{'a': 1, 'b': 3}] = 999
            h = Harvester(fn3_fba_runner, fl_pth, full_ds=mod_ds)
            h.save_full_ds()
            assert not h.full_ds.equals(fn3_fba_ds)
            h.harvest_cases([(1, 3)], overwrite=True)
            assert h.full_ds.equals(fn3_fba_ds)

    @pytest.mark.parametrize('engine', ['h5netcdf', 'netcdf4', 'zarr'])
    def test_harvest_cases_merge_dask(self, fn3_fba_runner,
                                      fn3_fba_ds, engine):
        import dask
        dask.config.set(scheduler='threads')

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth, engine=engine)
            h.harvest_cases([(1, 3), (2, 4)], chunks=1)
            h.harvest_cases([(1, 4), (2, 3)], chunks=1)
            assert not h.last_ds.identical(fn3_fba_ds)
            assert h.full_ds.identical(fn3_fba_ds)
            assert h.full_ds['sum'].chunks is not None
            h.full_ds.close()

    def test_harvest_cases_merge_dask_default(self, fn3_fba_runner,
                                              fn3_fba_ds):

        import dask
        dask.config.set(scheduler='threads')

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.h5')
            h = Harvester(fn3_fba_runner, fl_pth, engine='netcdf4',
                          chunks=1)
            h.harvest_cases([(1, 3), (2, 4)])
            h.harvest_cases([(1, 4), (2, 3)])
            assert not h.last_ds.identical(fn3_fba_ds)
            assert h.full_ds.identical(fn3_fba_ds)
            assert h.full_ds['sum'].chunks is not None
            h.full_ds.close()

    def test_expand_dims_and_ellipsis_combos(self):

        @label('x', harvester=True)
        def foo(a, b, c=5):
            return a + b + c

        foo.harvest_combos({
            'a': [1, 2],
            'b': [3, 4],
        })

        assert foo.full_ds['x'].ndim == 2
        foo.expand_dims('c', 5)
        assert foo.full_ds['x'].ndim == 3

        foo.harvest_combos({
            'a': ...,
            'b': ...,
            'c': [8, 9],
        })
        assert list(foo.full_ds['c'].values) == [5, 8, 9]

    def test_drop_sel(self):
        @label('x', harvester=True)
        def foo(a, b, c):
            return a + b + c

        foo.harvest_combos({
            'a': [1, 2],
            'b': [3, 4],
            'c': [5, 6, 7],
        })

        assert foo.full_ds['x'].ndim == 3
        assert foo.full_ds['x'].size == 12
        foo.drop_sel(c=6)
        assert foo.full_ds['x'].ndim == 3
        assert foo.full_ds['x'].size == 8
        assert 6 not in foo.full_ds.coords['c'].values


class TestSampler:

    @pytest.mark.parametrize("fname,engine", [
        ('test.pkl', 'pickle'),
        ('test.csv', 'csv'),
    ])
    def test_sample_combos_new(self, fname, engine):

        @label(var_names=['sum', 'diff', 'divisor', 'const'],
               constants={'c': 42})
        def sum_diff(a, b, c):
            return a + b, a - b, a % b == 0, c

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, fname)
            s = Sampler(sum_diff, fl_pth, engine=engine)
            print(s)
            combos = (
                ('a', (1, 2, 3, 4, 5)),
                ('b', lambda: np.random.randint(10, 20))
            )
            s.sample_combos(10, combos)
            hdf = load_df(fl_pth, engine=engine)

            assert len(hdf) == 10
            assert s.last_df.compare(hdf).empty
            assert s.full_df.compare(hdf).empty

            s.sample_combos(20, combos)
            hdf = load_df(fl_pth, engine=engine)
            assert len(hdf) == 30
            assert not s.last_df.equals(hdf)
            assert s.full_df.equals(hdf)

            assert set(s.full_df['const']) == set(s.full_df['c']) == {42}

            for col in ['sum', 'diff', 'divisor', 'const', 'a', 'b', 'c']:
                assert col in s.full_df

    @pytest.mark.parametrize('batchsize', [1, 3])
    def test_sow_reap_samples(self, batchsize):

        @label(var_names=['sum', 'diff', 'divisor', 'const'],
               constants={'c': 42})
        def sum_diff(a, b, c):
            return a + b, a - b, a % b == 0, c

        default_combos = (
            ('a', (1, 2, 3, 4, 5)),
            ('b', lambda: np.random.randint(10, 20))
        )

        with tempfile.TemporaryDirectory() as tmpdir:
            fl_pth = os.path.join(tmpdir, 'test.pkl')
            s = Sampler(sum_diff, fl_pth, default_combos=default_combos)
            c = s.Crop('test-crop-def', batchsize=batchsize, parent_dir=tmpdir)
            c.sow_samples(10)
            c.grow_missing()
            c.reap()
            hdf = load_df(fl_pth)
            assert len(hdf) == 10
            assert s.last_df.compare(hdf).empty
            assert s.full_df.compare(hdf).empty



import os
import tempfile

from pytest import fixture, mark, param, raises
import numpy as np
import xarray as xr

from xyzpy.manage import (
    load_ds,
    save_ds,
    save_merge_ds,
)


@fixture
def ds1():
    return xr.Dataset(
        coords={
            'b': ['l1', 'l2', 'l4'],
            'a': [1, 2, 3, 4]},
        data_vars={
            'x': (('b', 'a'),
                  np.random.randn(3, 4) + 1.0j * np.random.randn(3, 4)),
            'isodd': ('a', np.asarray([True, False, True, False]))},
        attrs={'foo': 'bar'})


@fixture
def ds2():
    return xr.Dataset(
        coords={
            'b': ['l3', 'l5'],
            'a': [3, 4, 5]},
        data_vars={
            'x': (('b', 'a'),
                  np.random.randn(2, 3) + 1.0j * np.random.randn(2, 3)),
            'isodd': ('a', np.asarray([True, False, True]))},
        attrs={'bar': 'baz'})


@fixture
def ds3():
    return xr.Dataset(
        coords={
            'b': ['l5'],
            'a': [4]},
        data_vars={
            'x': (('b', 'a'), [[123. + 456.0j]]),
            'isodd': ('a', np.asarray([True]))},
        attrs={'baz': 'qux'})


@fixture
def ds_real():
    return xr.Dataset(
        coords={
            'b': ['l3', 'l5'],
            'a': [3, 4, 5]},
        data_vars={
            'x': (('b', 'a'), np.random.randn(2, 3)),
            'isodd': ('a', np.asarray([True, False, True]))},
        attrs={'qux': 'corge'})


class TestSaveAndLoad:
    @mark.parametrize(("engine_save, engine_load"),
                      [('h5netcdf', 'h5netcdf'),
                       ('h5netcdf', 'netcdf4'),
                       ('zarr', 'zarr'),
                       ('joblib', 'joblib'),
                       param('netcdf4', 'h5netcdf', marks=mark.xfail),
                       param('netcdf4', 'netcdf4', marks=mark.xfail)])
    def test_io_only_real(self, ds_real, engine_save, engine_load):
        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds_real, os.path.join(tmpdir, "test.h5"),
                    engine=engine_save)
            ds2 = load_ds(os.path.join(tmpdir, "test.h5"), engine=engine_load)
            assert ds_real.equals(ds2)

    @mark.parametrize(("engine_save, engine_load"),
                      [('h5netcdf', 'h5netcdf'),
                       ('zarr', 'zarr'),
                       ('joblib', 'joblib'),
                       param('h5netcdf', 'netcdf4', marks=mark.xfail),
                       param('netcdf4', 'h5netcdf', marks=mark.xfail),
                       param('netcdf4', 'netcdf4', marks=mark.xfail)])
    def test_io_complex_data(self, ds1, engine_save, engine_load):
        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds1, os.path.join(tmpdir, "test.h5"), engine=engine_save)
            ds2 = load_ds(os.path.join(tmpdir, "test.h5"), engine=engine_load)
            assert ds1.identical(ds2)

    @mark.parametrize(("engine_load"),
                      ['h5netcdf', 'netcdf4'])
    def test_dask_load(self, ds_real, engine_load):
        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds_real, os.path.join(tmpdir, "test.nc"))
            ds2 = load_ds(os.path.join(tmpdir, "test.nc"),
                          engine=engine_load,
                          chunks=1)
            assert ds2.chunks['b'] == (1, 1)
            assert ds_real.identical(ds2)
            ds2.close()

    @mark.xfail
    def test_h5netcdf_dask(self, ds1):
        ds = ds1.chunk()
        with tempfile.TemporaryDirectory() as tmpdir:
            save_ds(ds, os.path.join(tmpdir, "test.h5"), engine='h5netcdf')
            ds2 = load_ds(os.path.join(tmpdir, "test.h5"))
            assert ds1.identical(ds2)
            ds2.close()

    def test_save_merge_ds(self, ds1, ds2, ds3):
        with tempfile.TemporaryDirectory() as tmpdir:
            fname = os.path.join(tmpdir, "test.h5")
            save_merge_ds(ds1, fname)
            save_merge_ds(ds2, fname)
            with raises(xr.MergeError):
                save_merge_ds(ds3, fname)
            save_merge_ds(ds3, fname, overwrite=True)
            exp = ds3.combine_first(xr.merge([ds1, ds2]))
            assert load_ds(fname).identical(exp)



import pytest
from xyzpy.gen.prepare import (
    parse_var_names,
    parse_var_dims,
)


class TestParsers:
    @pytest.mark.parametrize(
        " var_names,               expctd_var_names",
        [('x',                     ('x',)),
         (('x',),                  ('x',)),
         (('x', 'y'),              ('x', 'y')),
         (['x', 'y'],              ('x', 'y')),
         ((a for a in ('x', 'y')), ('x', 'y'))])
    def testparse_var_names(self, var_names, expctd_var_names):
        var_names = parse_var_names(var_names)
        assert var_names == expctd_var_names

    @pytest.mark.parametrize(
        " var_names,  var_dims,            expctd_var_dims",
        [('x',        None,                {'x': ()}),
         ('x',        [],                  {'x': ()}),
         ('x',        't',                 {'x': ('t',)}),
         ('x',        ['t'],               {'x': ('t',)}),
         ('x',        {'x': 't'},          {'x': ('t',)}),
         ('x',        {'x': ['t']},        {'x': ('t',)}),
         ('x',        [('x', ['t'])],      {'x': ('t',)}),
         (['x', 'y'], None,                {'x': (), 'y': ()}),
         (['x', 'y'], ['t', []],           {'x': ('t',), 'y': ()}),
         (['x', 'y'], {'x': 't'},          {'x': ('t',), 'y': ()}),
         (['x', 'y'], {('x', 'y'): 't'},   {'x': ('t',), 'y': ('t',)}),
         (['x', 'y'], {('x', 'y'): ['t']}, {'x': ('t',), 'y': ('t',)}),
         (['x', 'y'], ['t', 's'],          {'x': ('t',), 'y': ('s',)})])
    def test_parse_var_dims(self, var_names, var_dims, expctd_var_dims):
        var_dims = parse_var_dims(var_dims, var_names)
        assert var_dims == expctd_var_dims

    @pytest.mark.parametrize(
        " var_names,    var_dims",
        [(['x', 'y'], [['t', 's']]),
         (['x', 'y'], 't'),
         (['x', 'y'], {'z': 't'})])
    def test_var_dims_raises(self, var_names, var_dims):
        with pytest.raises(ValueError):
            parse_var_dims(var_dims, var_names)



from pytest import fixture, mark
import numpy as np
import xarray as xr
import matplotlib

from xyzpy import (
    lineplot,
    auto_lineplot,
    ilineplot,
    auto_ilineplot,
    visualize_matrix
)
from xyzpy.plot.color import convert_colors


matplotlib.use('Template')


@fixture
def dataset_3d():
    x = [1, 2, 3, 4, 5, 6, 8]
    z = [10, 20, 40, 80]
    d = np.random.rand(7, 4)
    ds = xr.Dataset()
    ds["x"] = x
    ds["z"] = z
    ds["y"] = (("x", "z"), d)
    return ds


@fixture
def dataset_heatmap():
    x = np.linspace(10, 20, 11)
    y = np.linspace(20, 40, 21)
    xx, yy = np.meshgrid(x, y)
    c = np.cos(((xx**2 + yy**2)**0.5) / 2)
    s = np.sin(((xx**2 + yy**2)**0.5) / 2)
    ds = xr.Dataset(
        coords={'x': x,
                'y': y},
        data_vars={'c': (('y', 'x'), c),
                   's': (('y', 'x'), s)})
    return ds.where(ds.y < 35)


@fixture
def dataset_4d():
    x = np.linspace(10, 20, 11)
    y = np.linspace(20, 40, 21)
    phi = np.linspace(-0.1, 0.1, 3)
    xx, yy, phis = np.meshgrid(x, y, phi)
    c = np.cos(((xx**2 + yy**2)**0.5) / 2 + phis)
    s = np.sin(((xx**2 + yy**2)**0.5) / 2 + phis)
    ds = xr.Dataset(
        coords={'x': x,
                'y': y,
                'phi': phi},
        data_vars={'c': (('y', 'x', 'phi'), c),
                   's': (('y', 'x', 'phi'), s)})
    return ds


@fixture
def dataset_5d():
    x = np.linspace(10, 20, 11)
    y = np.linspace(20, 40, 21)
    phi = np.linspace(-0.5, 0.5, 3)
    A = [1, 2]
    xx, yy, phis, AA = np.meshgrid(x, y, phi, A)
    c = AA * np.cos(((xx**2 + yy**2)**0.5) / 2 + phis)
    s = AA * np.sin(((xx**2 + yy**2)**0.5) / 2 + phis)
    ds = xr.Dataset(
        coords={'x': x,
                'y': y,
                'phi': phi,
                'A': A},
        data_vars={'c': (('y', 'x', 'phi', 'A'), c),
                   's': (('y', 'x', 'phi', 'A'), s)})
    return ds


@fixture
def dataset_scatter():
    x = np.random.randn(100, 100)
    y = np.random.randn(100, 100)
    z = np.random.randn(100)

    a = np.linspace(0, 3, 100)
    b = np.linspace(0, 3, 100)

    ds = xr.Dataset(
        coords={'a': a,
                'b': b},
        data_vars={'x': (('a', 'b'), x),
                   'y': (('a', 'b'), y),
                   'z': ('b', z)})
    return ds.where(ds.z > 0.5)


# --------------------------------------------------------------------------- #
# TEST COLORS                                                                 #
# --------------------------------------------------------------------------- #

class TestConvertColors:
    def test_simple(self):
        cols = [(1, 0, 0, 1), (0, 0.5, 0, 0.5)]
        new_cols = list(convert_colors(cols, outformat='BOKEH'))
        assert new_cols == [(255, 0, 0, 1), (0, 127, 0, 0.5)]


# --------------------------------------------------------------------------- #
# TEST PLOTTERS                                                               #
# --------------------------------------------------------------------------- #

@mark.parametrize("plot_fn", [lineplot, ilineplot])
class TestCommonInterface:
    @mark.parametrize("colors", [True, False, None])
    @mark.parametrize("markers", [True, False, None])
    def test_works_2d(self,
                      plot_fn,
                      dataset_3d,
                      colors,
                      markers):
        plot_fn(dataset_3d, "x", "y", "z", return_fig=True,
                colors=colors,
                markers=markers)

    @mark.parametrize("colormap", ['xyz', 'viridis'])
    @mark.parametrize("colormap_log", [True, False])
    @mark.parametrize("colormap_reverse", [True, False])
    @mark.parametrize("string_z_coo", [True, False])
    def test_color_options(self,
                           plot_fn,
                           dataset_3d,
                           string_z_coo,
                           colormap,
                           colormap_log,
                           colormap_reverse):
        if string_z_coo:
            dataset_3d = dataset_3d.copy(deep=True)
            dataset_3d['z'] = ['a', 'b', 'c', 'd']
        plot_fn(dataset_3d, "x", "y", "z", return_fig=True,
                colors=True,
                colormap=colormap,
                colormap_log=colormap_log,
                colormap_reverse=colormap_reverse)

    @mark.parametrize("markers", [True, False, None])
    def test_works_1d(self,
                      plot_fn,
                      dataset_3d,
                      markers):
        plot_fn(dataset_3d.loc[{"z": 40}], "x", "y", return_fig=True,
                markers=markers)

    @mark.parametrize("padding", [None, 0.1])
    @mark.parametrize("xlims", [None, (0., 10.)])
    @mark.parametrize("ylims", [None, (-1., 1.)])
    def test_plot_range(self,
                        dataset_3d,
                        plot_fn,
                        padding,
                        xlims,
                        ylims):
        plot_fn(dataset_3d, "x", "y", "z", return_fig=True,
                padding=padding,
                xlims=xlims,
                ylims=ylims)

    @mark.parametrize("xlog", [True, False])
    @mark.parametrize("ylog", [True, False])
    @mark.parametrize("xticks", [None, (2, 3,)])
    @mark.parametrize("yticks", [None, (0.2, 0.3,)])
    @mark.parametrize("vlines", [None, (2, 3,)])
    @mark.parametrize("hlines", [None, (0.2, 0.3,)])
    def test_ticks_and_lines(self,
                             dataset_3d,
                             plot_fn,
                             xlog,
                             ylog,
                             xticks,
                             yticks,
                             vlines,
                             hlines):
        plot_fn(dataset_3d, "x", "y", "z", return_fig=True,
                xticks=xticks,
                yticks=yticks,
                vlines=vlines,
                hlines=hlines)


class TestLinePlot:

    def test_auto_lineplot(self):
        x = [1, 2, 3]
        y = [[4, 5, 6], [7, 8, 9]]
        auto_lineplot(x, y)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.lineplot('x', 'c', 'y', row='phi')
        dataset_4d.xyz.lineplot('x', 'c', 'y', col='phi')

    @mark.parametrize("colors", [False, True])
    def test_multi_plot_5d(self, dataset_5d, colors):
        kws = {'colors': colors}
        dataset_5d.xyz.lineplot('x', 'c', 'y', row='phi', col='A', **kws)
        dataset_5d.xyz.lineplot('x', 'c', 'y', col='phi', row='A', **kws)


class TestILinePlot:

    def test_auto_ilineplot(self):
        x = [1, 2, 3]
        y = [[4, 5, 6], [7, 8, 9]]
        auto_ilineplot(x, y)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.ilineplot('x', 'c', 'y', row='phi')
        dataset_4d.xyz.ilineplot('x', 'c', 'y', col='phi')

    @mark.parametrize("colors", [False, True])
    def test_multi_plot_5d(self, dataset_5d, colors):
        kws = {'colors': colors}
        dataset_5d.xyz.ilineplot('x', 'c', 'y', row='phi', col='A', **kws)
        dataset_5d.xyz.ilineplot('x', 'c', 'y', col='phi', row='A', **kws)


class TestHeatmap:
    def test_simple(self, dataset_heatmap):
        dataset_heatmap.xyz.heatmap('x', 'y', 'c', return_fig=True)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.heatmap('x', 'y', 'c', row='phi')
        dataset_4d.xyz.heatmap('x', 'y', 'c', col='phi')

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.heatmap('x', 'y', 'c', row='phi', col='A')
        dataset_5d.xyz.heatmap('x', 'y', 'c', col='phi', row='A')


class TestIHeatmap:
    def test_simple(self, dataset_heatmap):
        dataset_heatmap.xyz.iheatmap('x', 'y', 'c', return_fig=True)

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.iheatmap('x', 'y', 'c', row='phi')
        dataset_4d.xyz.iheatmap('x', 'y', 'c', col='phi')

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.iheatmap('x', 'y', 'c', row='phi', col='A')
        dataset_5d.xyz.iheatmap('x', 'y', 'c', col='phi', row='A')


class TestScatter:

    def test_normal(self, dataset_scatter):
        dataset_scatter.xyz.scatter('x', 'y')

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.scatter('x', 'y', c='c', row='phi')
        dataset_4d.xyz.scatter('x', 'y', c='c', col='phi')

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.scatter('x', 'y', c='c', row='phi', col='A')
        dataset_5d.xyz.scatter('x', 'y', c='c', col='phi', row='A')


class TestIScatter:

    def test_normal(self, dataset_scatter):
        dataset_scatter.xyz.iscatter('x', 'y')

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.iscatter('x', 'y', c='c', row='phi')
        dataset_4d.xyz.iscatter('x', 'y', c='c', col='phi')

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.iscatter('x', 'y', c='c', row='phi', col='A')
        dataset_5d.xyz.iscatter('x', 'y', c='c', col='phi', row='A')


class TestHistogram:

    def test_normal(self, dataset_3d):
        dataset_3d.xyz.histogram('y', z='z')

    def test_multi_hist(self, dataset_heatmap):
        dataset_heatmap.xyz.histogram(('c', 's'))

    def test_multi_plot_4d(self, dataset_4d):
        dataset_4d.xyz.histogram(('c', 's'), row='phi')
        dataset_4d.xyz.histogram(('c', 's'), col='phi')

    def test_multi_plot_5d(self, dataset_5d):
        dataset_5d.xyz.histogram(('c', 's'), row='phi', col='A')
        dataset_5d.xyz.histogram(('c', 's'), col='phi', row='A')


class TestVisualizeMatrix():

    def test_single(self):
        x = np.random.randn(10, 10)
        visualize_matrix(x)



"""Functions for systematically evaluating a function over specific cases.
"""
import functools
import itertools

from ..utils import progbar
from .prepare import (
    parse_fn_args,
    parse_cases,
    parse_combos,
    parse_constants,
    parse_resources,
    parse_var_names,
    parse_var_dims,
    parse_var_coords,
)
from .combo_runner import (
    combo_runner_core,
    combo_runner_to_ds,
)


def case_runner(
    fn,
    fn_args,
    cases,
    combos=None,
    constants=None,
    split=False,
    shuffle=False,
    parse=True,
    parallel=False,
    executor=None,
    num_workers=None,
    verbosity=1,
):
    """Simple case runner that outputs the raw tuple of results.

    Parameters
    ----------
    fn : callable
        Function with which to evalute cases with
    fn_args : tuple
        Names of case arguments that fn takes, can be ``None`` if each case is
        a ``dict``.
    cases : iterable[tuple] or iterable[dict]
        List of specific configurations that ``fn_args`` should take. If
        ``fn_args`` is ``None``, each case should be a ``dict``.
    combos : dict_like[str, iterable], optional
        Optional specification of sub-combinations.
    constants : dict, optional
        Constant function arguments.
    split : bool, optional
        See :func:`~xyzpy.combo_runner`.
    shuffle : bool or int, optional
        If given, compute the results in a random order (using ``random.seed``
        and ``random.shuffle``), which can be helpful for distributing
        resources when not all cases are computationally equal.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    executor : executor-like pool, optional
        Submit all combos to this pool executor. Must have ``submit`` or
        ``apply_async`` methods and API matching either ``concurrent.futures``
        or an ``ipyparallel`` view. Pools from ``multiprocessing.pool`` are
        also  supported.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

            - 0: nothing,
            - 1: just progress,
            - 2: all information.

    Returns
    -------
        results : list of fn output for each case
    """
    if parse:
        # Prepare fn_args and values
        fn_args = parse_fn_args(fn, fn_args)
        cases = parse_cases(cases, fn_args)
        combos = parse_combos(combos)
        constants = parse_constants(constants)

    return combo_runner_core(
        fn,
        cases=cases,
        combos=combos,
        constants=constants,
        parallel=parallel,
        num_workers=num_workers,
        executor=executor,
        verbosity=verbosity,
        split=split,
        flat=True,
        shuffle=shuffle,
    )


def case_runner_to_ds(
    fn,
    fn_args,
    cases,
    var_names,
    var_dims=None,
    var_coords=None,
    combos=None,
    constants=None,
    resources=None,
    attrs=None,
    shuffle=False,
    to_df=False,
    parse=True,
    parallel=False,
    num_workers=None,
    executor=None,
    verbosity=1,
):
    """Takes a list of ``cases`` to run ``fn`` over, possibly in parallel, and
    outputs a :class:`xarray.Dataset`.

    Parameters
    ----------
    fn : callable
        Function to evaluate.
    fn_args : str or iterable[str]
        Names and order of arguments to ``fn``, can be ``None`` if ``cases``
        are supplied as dicts.
    cases: iterable[tuple] or iterable[dict]
        List of configurations used to generate results.
    var_names : str or iterable of str
        Variable name(s) of the output(s) of ``fn``.
    var_dims : sequence of either strings or string sequences, optional
        'Internal' names of dimensions for each variable, the values for
        each dimension should be contained as a mapping in either
        `var_coords` (not needed by `fn`) or `constants` (needed by `fn`).
    var_coords : mapping, optional
        Mapping of extra coords the output variables may depend on.
    combos : iterable[tuple] or iterable[dict], optional
        If specified, run all combinations of some arguments in these mappings.
    constants : mapping, optional
        Arguments to `fn` which are not iterated over, these will be
        recorded either as attributes or coordinates if they are named
        in `var_dims`.
    resources : mapping, optional
        Like `constants` but they will not be recorded.
    attrs : mapping, optional
        Any extra attributes to store.
    shuffle : bool or int, optional
        If given, compute the results in a random order (using ``random.seed``
        and ``random.shuffle``), which can be helpful for distributing
        resources when not all cases are computationally equal.
    parse : bool, optional
        Whether to perform parsing of the inputs arguments.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    executor : executor-like pool, optional
        Submit all combos to this pool executor. Must have ``submit`` or
        ``apply_async`` methods and API matching either ``concurrent.futures``
        or an ``ipyparallel`` view. Pools from ``multiprocessing.pool`` are
        also  supported.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

            - 0: nothing,
            - 1: just progress,
            - 2: all information.

    Returns
    -------
    ds : xarray.Dataset
        Dataset with minimal covering coordinates and all cases
        evaluated.
    """
    if parse:
        # Prepare fn_args and values
        fn_args = parse_fn_args(fn, fn_args)
        cases = parse_cases(cases, fn_args)
        combos = parse_combos(combos)
        constants = parse_constants(constants)
        resources = parse_resources(resources)
        var_names = parse_var_names(var_names)
        var_dims = parse_var_dims(var_dims, var_names=var_names)
        var_coords = parse_var_coords(var_coords)

    return combo_runner_to_ds(
        fn=fn,
        combos=combos,
        var_names=var_names,
        var_dims=var_dims,
        var_coords=var_coords,
        cases=cases,
        constants=constants,
        resources=resources,
        attrs=attrs,
        shuffle=shuffle,
        to_df=to_df,
        parallel=parallel,
        num_workers=num_workers,
        executor=executor,
        verbosity=verbosity,
        parse=False,
    )


case_runner_to_df = functools.partial(case_runner_to_ds, to_df=True)


# --------------------------------------------------------------------------- #
# Update or add new values                                                    #
# --------------------------------------------------------------------------- #

def find_missing_cases(ds, ignore_dims=None, show_progbar=False):
    """Find all cases in a dataset with missing data.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset in which to find missing data
    ignore_dims : set (optional)
        internal variable dimensions (i.e. to ignore)
    show_progbar : bool (optional)
        Show the current progress

    Returns
    -------
    missing_fn_args, missing_cases :
        Function arguments and missing cases.
    """
    # Parse ignore_dims
    ignore_dims = ({ignore_dims} if isinstance(ignore_dims, str) else
                   set(ignore_dims) if ignore_dims else set())

    # Find all configurations
    fn_args = tuple(coo for coo in ds.dims if coo not in ignore_dims)
    var_names = tuple(ds.data_vars)
    all_cases = itertools.product(*(ds[arg].data for arg in fn_args))

    # Only return those corresponding to all missing data
    def gen_missing_list():
        for case in progbar(all_cases, disable=not show_progbar):
            sub_ds = ds.loc[dict(zip(fn_args, case))]
            if all(sub_ds[v].isnull().all() for v in var_names):
                yield case

    return fn_args, tuple(gen_missing_list())
